/*
 * Program: quasijarus2gzip
 * File: quasijarus2gzip.c
 * Created: 2023-04-17 21:55:39
 * Author: Lucio F. Albenga <lucio@albenga.es>
 * Description: Converts a file compressed in Quasijarus Strong Compressed Data
 *              to a file compressed in gzip format
 *
 * Copyright (c) 2023 Lucio F. Albenga
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation version 3 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: AGPL-3.0-only
 */


#include <stdbool.h>
#include <stdio.h>
#include <string.h>

#define MAGIC_ONE 0x1f  /* Magic number one is the same for Quasijarus and Gzip */
#define GZIP_MAGIC_TWO 0x8b
#define QUASIJARUS_MAGIC_TWO 0xa1
#define BUFFER_SIZE 1024 

/*
 * Check out RFC1952 in order to know the GZIP file format
 * https://datatracker.ietf.org/doc/html/rfc1952
 */
struct gzip_header {
        unsigned char magic_1;
        unsigned char magic_2;
        unsigned char compression_method;
        unsigned char header_flags;
        unsigned char timestamp[4];
        unsigned char compression_flags;
        unsigned char operating_system_id;
};


void print_usage();
int convert_file(char *input, char *output);
FILE* open_file(char *filename, char *mode);
bool is_quasijarus(int chars_read, unsigned char *magic_number);
void fill_gzip_header(struct gzip_header *gz_header);
void write_file(FILE *fp_in, FILE *fp_out);

int main(int argc, char *argv[])
{
        if ((argc == 2) &&
	    (strcmp(argv[1], "--help") == 0 || strcmp(argv[1], "-h") == 0)) {
		print_usage();
		return 0;
	} else if (argc != 3) {
                puts("\nERROR: wrong number of arguments\n");
		print_usage();
                return 1;
        } else {
		char *input_filename = argv[1];
		char *output_filename = argv[2];
		return convert_file(input_filename, output_filename);
	}
}

void print_usage()
{
        printf("Usage: quasijarus2gzip <input_file.Z> <output_file.gz>\n\n");
}

int convert_file(char *input, char *output)
{
	FILE *fp_in, *fp_out;
        unsigned char magic_number[2] = {'\0'};
        int chars_read = 0;

        fp_in = open_file(input, "rb");
	if (fp_in == NULL) {
		return 1;
	}
	
        chars_read = fread(&magic_number, sizeof(char), 2, fp_in);
	if (!is_quasijarus(chars_read, magic_number)) {
		fclose(fp_in);
		printf("Error file '%s' is not a Quasijarus Strong Compressed "
		       "Data file\n",
		       output);
		return 1;
	}
	
        fp_out = fopen(output, "wb");
        if (fp_out == NULL) {
		fclose(fp_in);
                return 1;
        }
	
        write_file(fp_in, fp_out);
        fclose(fp_out);
        fclose(fp_in);
        return 0;
}

FILE* open_file(char *filename, char *mode)
{
	FILE *fp = fopen(filename, mode);
        if (fp == NULL) {
                printf("Can't open file: %s\n", filename);
        }
	return fp;
}

bool is_quasijarus(int chars_read, unsigned char *magic_number)
{
	return ( (chars_read == 2) &&
		 magic_number[0] == MAGIC_ONE &&
		 magic_number[1] == QUASIJARUS_MAGIC_TWO);
}

void write_file(FILE *fp_in, FILE *fp_out)
{
	int chars_read = 0;
	struct gzip_header gz_header;
	unsigned char buffer[BUFFER_SIZE];
        fill_gzip_header(&gz_header);
        fwrite(&gz_header, sizeof(struct gzip_header), 1, fp_out);
        do {
                chars_read = fread(buffer, sizeof(char), BUFFER_SIZE, fp_in);
                if (chars_read >= 0) {
                        fwrite(buffer, sizeof(char), chars_read, fp_out);
                }
                
        } while(chars_read > 0);

}

void fill_gzip_header(struct gzip_header *gz_header)
{
        gz_header->magic_1 = MAGIC_ONE;
        gz_header->magic_2 = GZIP_MAGIC_TWO;
        gz_header->compression_method = 0x08; /* deflate */
        gz_header->header_flags = 0;
        gz_header->timestamp[0] = 0;
        gz_header->timestamp[1] = 0;
        gz_header->timestamp[2] = 0;
        gz_header->timestamp[3] = 0;
        gz_header->compression_flags = 0;
        gz_header->operating_system_id = 0x03; /* Unix */
}
